# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common HTML controls."""

# pylint: disable=g-importing-member
# pylint: disable=g-bad-import-order

from pyglove.core.views.html.controls.base import HtmlControl


from pyglove.core.views.html.controls.label import Label
from pyglove.core.views.html.controls.label import LabelGroup
from pyglove.core.views.html.controls.label import Badge

from pyglove.core.views.html.controls.tooltip import Tooltip

from pyglove.core.views.html.controls.tab import Tab
from pyglove.core.views.html.controls.tab import TabControl

from pyglove.core.views.html.controls.progress_bar import ProgressBar
from pyglove.core.views.html.controls.progress_bar import SubProgress

# pylint: enable=g-bad-import-order
# pylint: enable=g-importing-member



# Copyright 2025 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Injecting and manipulating values through context managers."""

import contextlib
import dataclasses
import threading
from typing import Any, Callable, ContextManager, Iterator, Optional

from pyglove.core.utils import missing

RAISE_IF_HAS_ERROR = (missing.MISSING_VALUE,)
_TLS_KEY_CONTEXTUAL_OVERRIDES = '__contextual_overrides__'
_global_contextual_overrides = threading.local()


@dataclasses.dataclass(frozen=True)
class ContextualOverride:
  """Value marker for contextual override for an attribute."""

  # Overridden value.
  value: Any

  # If True, this override will apply to both current scope and nested scope,
  # meaning current `pg.contextual_override` will take precedence over all
  # nested `pg.contextual_override` on this attribute.
  cascade: bool = False

  # If True, this override will apply to attributes that already have values.
  override_attrs: bool = False


def contextual_override(
    *,
    cascade: bool = False,
    override_attrs: bool = False,
    **variables,
) -> ContextManager[dict[str, ContextualOverride]]:
  """Context manager to provide contextual values under a scope.

  Please be aware that contextual value override are per-thread. If you want
  to propagate the contextual value override to other threads, please obtain
  a wrapper function for a user function using
  `pg.with_contextual_override(func)`.

  Args:
    cascade: If True, this override will apply to both current scope and nested
      scope, meaning that this `pg.contextual_override` will take precedence
      over all nested `pg.contextual_override` on the overriden variables.
    override_attrs: If True, this override will apply to attributes that already
      have values. Otherwise overridden variables will only be used for
      contextual attributes whose values are not present.
    **variables: Key/values as override for contextual attributes.

  Returns:
    A dict of attribute names to their contextual overrides.
  """
  vs = {}
  for k, v in variables.items():
    if not isinstance(v, ContextualOverride):
      v = ContextualOverride(v, cascade, override_attrs)
    vs[k] = v
  return contextual_scope(_global_contextual_overrides, **vs)


def with_contextual_override(func: Callable[..., Any]) -> Callable[..., Any]:
  """Wraps a user function with the access to the current contextual override.

  The wrapped function can be called from another thread.

  Args:
    func: The user function to be wrapped.

  Returns:
    A wrapper function that have the access to the current contextual override,
    which can be called from another thread.
  """
  with contextual_override() as current_context:
    pass

  def _func(*args, **kwargs) -> Any:
    with contextual_override(**current_context):
      return func(*args, **kwargs)

  return _func


def get_contextual_override(var_name: str) -> Optional[ContextualOverride]:
  """Returns the overriden contextual value in current scope."""
  return get_scoped_value(_global_contextual_overrides, var_name)


def contextual_value(var_name: str, default: Any = RAISE_IF_HAS_ERROR) -> Any:
  """Returns the value of a variable defined in `pg.contextual_override`."""
  override = get_contextual_override(var_name)
  if override is None:
    if default == RAISE_IF_HAS_ERROR:
      raise KeyError(f'{var_name!r} does not exist in current context.')
    return default
  return override.value


def all_contextual_values() -> dict[str, Any]:
  """Returns all values provided from `pg.contextual_override` in scope."""
  overrides = getattr(
      _global_contextual_overrides, _TLS_KEY_CONTEXTUAL_OVERRIDES, {}
  )
  return {k: v.value for k, v in overrides.items()}


@contextlib.contextmanager
def contextual_scope(
    tls: threading.local, **variables
) -> Iterator[dict[str, ContextualOverride]]:
  """Context manager to set variables within a scope."""
  previous_values = getattr(tls, _TLS_KEY_CONTEXTUAL_OVERRIDES, {})
  current_values = dict(previous_values)
  for k, v in variables.items():
    old_v = current_values.get(k, None)
    if old_v and old_v.cascade:
      v = old_v
    current_values[k] = v
  try:
    setattr(tls, _TLS_KEY_CONTEXTUAL_OVERRIDES, current_values)
    yield current_values
  finally:
    setattr(tls, _TLS_KEY_CONTEXTUAL_OVERRIDES, previous_values)


def get_scoped_value(
    tls: threading.local, var_name: str, default: Any = None
) -> ContextualOverride:
  """Gets the value for requested variable from current scope."""
  scoped_values = getattr(tls, _TLS_KEY_CONTEXTUAL_OVERRIDES, {})
  return scoped_values.get(var_name, default)




# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import unittest

from pyglove.core import symbolic   # pylint: disable=unused-import
from pyglove.core.views.html.controls import label as label_lib


class TestCase(unittest.TestCase):

  def assert_html_content(self, control, expected):
    expected = inspect.cleandoc(expected).strip()
    actual = control.to_html().content.strip()
    if actual != expected:
      print(actual)
    self.assertEqual(actual, expected)


class LabelTest(TestCase):

  def test_text_only(self):
    label = label_lib.Label('foo')
    self.assertIsNone(label.tooltip)
    self.assertIsNone(label.link)
    self.assert_html_content(label, '<span class="label">foo</span>')
    with self.assertRaisesRegex(ValueError, 'Non-interactive .*'):
      label.update('bar')

  def test_with_link(self):
    label = label_lib.Label('foo', link='http://google.com', id='my-foo')
    self.assertEqual(label.element_id(), 'my-foo')
    self.assertIsNone(label.tooltip)
    self.assertEqual(label.link, 'http://google.com')
    self.assert_html_content(
        label, '<a class="label" id="my-foo" href="http://google.com">foo</a>'
    )

  def test_with_tooltip(self):
    label = label_lib.Label('foo', tooltip='bar')
    self.assertEqual(label.tooltip.for_element, '.label')
    self.assertIsNone(label.link)
    self.assert_html_content(
        label,
        (
            '<div class="label-container"><span class="label">foo</span>'
            '<span class="tooltip">bar</span></div>'
        )
    )

  def test_update(self):
    label = label_lib.Label(
        'foo', 'bar', 'http://google.com',
        interactive=True,
        css_classes=['foo', 'bar'],
    )
    self.assertIn('id="control-', label.to_html_str(content_only=True))
    with label.track_scripts() as scripts:
      label.update(
          'bar',
          tooltip='baz',
          link='http://www.yahoo.com',
          styles=dict(color='red'),
          add_class=['baz'],
          remove_class=['bar'],
      )
    self.assertEqual(label.text, 'bar')
    self.assertEqual(label.tooltip.content, 'baz')
    self.assertEqual(label.link, 'http://www.yahoo.com')
    self.assertEqual(label.styles, dict(color='red'))
    self.assertEqual(label.css_classes, ['foo', 'baz'])
    self.assertEqual(
        scripts,
        [
            inspect.cleandoc(
                f"""
                elem = document.getElementById("{label.element_id()}");
                elem.textContent = "bar";
                """
            ),
            inspect.cleandoc(
                f"""
                elem = document.getElementById("{label.element_id()}");
                elem.style = "color:red;";
                """
            ),
            inspect.cleandoc(
                f"""
                elem = document.getElementById("{label.element_id()}");
                elem.href = "http://www.yahoo.com";
                """
            ),
            inspect.cleandoc(
                f"""
                elem = document.getElementById("{label.tooltip.element_id()}");
                elem.textContent = "baz";
                """
            ),
            inspect.cleandoc(
                f"""
                elem = document.getElementById("{label.tooltip.element_id()}");
                elem.classList.remove("html-content");
                """
            ),
            inspect.cleandoc(
                f"""
                elem = document.getElementById("{label.element_id()}");
                elem.classList.add("baz");
                """
            ),
            inspect.cleandoc(
                f"""
                elem = document.getElementById("{label.element_id()}");
                elem.classList.remove("bar");
                """
            ),
        ]
    )

  def test_badge(self):
    badge = label_lib.Badge('foo')
    self.assert_html_content(badge, '<span class="label badge">foo</span>')


class LabelGroupTest(TestCase):

  def test_basic(self):
    group = label_lib.LabelGroup(['foo', 'bar'], name='baz')
    self.assert_html_content(
        group,
        (
            '<div class="label-group"><span class="label group-name">baz</span>'
            '<span class="label group-value">foo</span>'
            '<span class="label group-value">bar</span></div>'
        )
    )

  def test_interactive(self):
    group = label_lib.LabelGroup(['foo', 'bar'], name='baz', interactive=True)
    self.assertTrue(group.name.interactive)
    self.assertTrue(group.labels[0].interactive)
    self.assertTrue(group.labels[1].interactive)


if __name__ == '__main__':
  unittest.main()
